import logging

import spread_knx as knx
from ..entities import Funit
from ..providers import Provider


class KnxProvider(Provider):
    def_state = None

    def __init__(self, id: int, address: knx.address.Address, **kwargs):
        super(KnxProvider, self).__init__(id, **kwargs)
        self._address = address
        self._status = self.SUSPICIOUS

    @property
    def binding(self) -> knx.address.Address:
        return self._address

    def parse_bindings(self, data):
        pass

    def on_survey(self):
        yield from super(KnxProvider, self).on_survey()
        yield self.check_valid()

    def request(self, cmd, *args, sig=None, state_address=None, **kwargs):
        for response in self._manager.send_interface.send(cmd, state_address):
            self.on_request_answer(response, *args, sig=sig)

    def on_request_answer(self, response: knx.commands.DPTU, *args, sig=None, **kwargs):
        logging.warning('Unknown response %s' % response)

    def check_valid(self):
        raise NotImplementedError()

    def get_info(self):
        pass

    def prs2raw(self, p):
        if p == 0:
            return 0
        return min(max(1, round(255 * p / 100)), 255)

    def raw2prs(self, r):
        return None if r is None else min(100, round(100 * r / 255))

    def SetDiscovery(self, value, sig=None):
        pass

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        raise NotImplemented()

    def GetBrightnessLevel(self, sig=None):
        raise NotImplemented()


class KnxLights(KnxProvider):
    @staticmethod
    def parse_args(attributes, **kwargs) -> dict:
        return dict(
            **kwargs,
            address=knx.address.LightAddress.deserialize(attributes)
        )

    @property
    def binding(self) -> knx.address.LightAddress:
        return self._address

    def repair_on_fail(self) -> tuple:
        return (*super(KnxProvider, self).repair_on_fail(), BrightnessLevel,)

    def check_valid(self):
        self.GetBrightnessLevel()

        level = self.get_value(BrightnessLevel)
        self.set_valid(self.SUSPICIOUS if level is None else self.VALID)


class KnxDimmer(KnxLights):
    def_state = knx.commands.Level

    def on_request_answer(self, response: knx.commands.DPTU, *args, sig=None):
        if isinstance(response, knx.commands.Level):
            self.on_update(BrightnessLevel, self.raw2prs(response.value), sig=sig)

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        self.on_update(BrightnessLevel, value, preliminary=True, sig=sig)
        self.request(knx.commands.Level.set(self.binding.control, self.prs2raw(value)), sig=sig)

    def GetBrightnessLevel(self, sig=None):
        self.request(knx.commands.Level.get(self.binding.state), state_address=self.binding.state, sig=sig)


class KnxRelay(KnxLights):
    def_state = knx.commands.Switching

    def on_request_answer(self, response: knx.commands.DPTU, *args, sig=None):
        if isinstance(response, knx.commands.Switching):
            value = response.value
            if value is not None:
                value = 100 if value else 0
            self.on_update(BrightnessLevel, value, sig=sig)

    def SetBrightnessLevel(self, value, sig=None, **kwargs):
        self.on_update(BrightnessLevel, value, preliminary=True, sig=sig)
        self.request(knx.commands.Switching.set(self.binding.control, value > 0), sig=sig)

    def GetBrightnessLevel(self, sig=None):
        self.request(knx.commands.Switching.get(self.binding.state), state_address=self.binding.state, sig=sig)


class KnxMotor(KnxProvider):
    def_state = knx.commands.Level

    @staticmethod
    def parse_args(attributes, **kwargs) -> dict:
        return dict(
            **kwargs,
            address=knx.address.MotorAddress.deserialize(attributes)
        )

    @property
    def binding(self) -> knx.address.MotorAddress:
        return self._address

    def repair_on_fail(self) -> tuple:
        return (*super(KnxProvider, self).repair_on_fail(), PositionLevel,)

    def check_valid(self):
        self.GetLevel()

        level = self.get_value(PositionLevel)
        self.set_valid(self.SUSPICIOUS if level is None else self.VALID)

    def on_request_answer(self, response: knx.commands.DPTU, *args, sig=None):
        if isinstance(response, knx.commands.Level):
            self.on_update(PositionLevel, self.raw2prs(response.value), sig=sig)

    def SetLevel(self, value, sig=None, **kwargs):
        if value is None:
            self.request(knx.commands.Step.set(self.binding.stop, 1), sig=sig)
            # Значение возвращается само по завершении дижения
            # self.GetLevel()
        else:
            level = self.prs2raw(value)
            has_level = self.get_value(PositionLevel)

            self.on_update(PositionLevel, value, preliminary=True, sig=sig)

            if level == has_level:
                self.GetLevel(sig=sig)
            else:
                self.request(knx.commands.Level.set(self.binding.control, level), sig=sig)

    def GetLevel(self, sig=None):
        self.request(knx.commands.Level.get(self.binding.state), state_address=self.binding.state, sig=sig)


class KNXFunit(Funit):
    entity_type = KnxProvider
    easy_get = False
    get_after_set = False
    set_cmd = None
    get_cmd = None
    set_func = None
    get_func = None

    @property
    def entity(self) -> KnxProvider:
        return super(KNXFunit, self).entity

    def set(self):
        if self.set_cmd:
            self.entity.request(self.set_cmd.set(self.entity.binding.control, self.set_value), sig=self.signature)
        elif self.set_func:
            getattr(self.entity, self.set_func.__name__)(self.set_value, **self.params, sig=self.signature)

        if self.get_cmd and self.get_after_set:
            self.get()

    def get(self):
        if self.easy_get:
            self.entity.on_update(self.funit_type, self.entity.get_value(self.funit_type), self.signature)
        elif self.get_cmd:
            self.entity.request(self.get_cmd.get(self.entity.binding.state), sig=self.signature)
        elif self.get_func:
            getattr(self.entity, self.get_func.__name__)(**self.params, sig=self.signature)


KnxProvider.funit_class = KNXFunit


class KnxLightFunit(KNXFunit):
    pass


class BrightnessLevel(KnxLightFunit):
    set_func = KnxProvider.SetBrightnessLevel
    get_func = KnxProvider.GetBrightnessLevel
    value_range = range(0, 101)


class GroupLevel(BrightnessLevel):
    pass


class KnxMotorFunit(KNXFunit):
    pass


class PositionLevel(KnxMotorFunit):
    set_func = KnxMotor.SetLevel
    get_func = KnxMotor.GetLevel
    value_range = range(0, 101)
