from unittest import TestCase

from spread_core.bam.engineries import Enginery

from spread_core.bam.engineries.alarm import FireSensor, LeakageSensor, IntruderSensor

from spread_core.bam.subgineries import alarming as alarming
from spread_core.mqtt.spread.variable import Variable
from spread_core.tests.equipment.utils import TestEquipmentAdapter, TestEngineryRecipe


class TestAlarm(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.adapter = TestEquipmentAdapter()

    def setUp(self) -> None:
        ENGS = [FireSensor, IntruderSensor, LeakageSensor, Enginery]

        self.adapter.add_manager(2, "Manager")

        for i in range(12):
            e = ENGS[i//3](i)
            self.adapter.add_provider(TestEngineryRecipe(1000 + i))
            self.adapter.add_enginery(e, 1000+i, "Simple")

        self._sb = alarming.Alarm(99, 2, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])
        self._sb.set_adapter(self.adapter)

    def _check_publish_state(self, funit, value, delay=0, params=None):
        self.__check_publish('Spread/State/1/Location/99/Alarm/%s' % funit, value, _retain=True, _delay=delay, params=params)

    def _check_publish_command(self, p_id, funit, value, params=None):
        self.__check_publish('Spread/Set/Provider/%s/%s' % (p_id, funit), value, _retain=False, _delay=0, params=params)

    def _check_end_of_publishes(self):
        if self.adapter.queue:
            unexpected_publishes = f'Unexpected publishes({len(self.adapter.queue)} pcs):\n'
            unexpected_publishes += "\n".join(str(i) for i in self.adapter.queue)
            self.fail(unexpected_publishes)

    def __check_publish(self, _topic, _value, _retain=True, _delay=0, params=None):
        if len(self.adapter.queue) == 0:
            self.fail('Expected publish is absent: %s' % _topic)
        else:
            topic, variable, retain, delay = self.adapter.queue.popleft()
            self.assertEqual(_topic, topic)
            self.assertEqual(_value, variable.value)
            self.assertEqual(_retain, retain)
            self.assertEqual(_delay, delay)

            if params:
                for param_name in params:
                    try:
                        self.assertEqual(params[param_name], variable.params[param_name])
                    except KeyError:
                        self.fail('Expected variable parameter "%s" is absent' % param_name)

    @property
    def sb(self) -> alarming.Alarm:
        return self._sb

    def test_init(self):
        self.assertEqual(1, self.sb.project_id)
        self.assertEqual(2, self.sb.location_id)
        self.assertEqual(99, self.sb.id)
        self.assertEqual(self.sb._enginery_ids, [e.id for e in self.sb.engineries])

        for e in self.sb.engineries:
            assert isinstance(e, Enginery)
            self.assertIn(self.sb, e._subgineries)

    def test_address(self):
        self.assertEqual('1/Location/99/Alarm/SomeFunit', self.sb.get_address("SomeFunit"))

    def test_state_publishing(self):
        self.sb.publish_state('SomeFunit', 'SomeValue')
        self._check_publish_state('SomeFunit', 'SomeValue')
        self._check_end_of_publishes()

    def test_invalidation(self):
        self.sb.values['SomeFunit'] = 2
        self.sb.set_invalid()

        self._check_publish_state('SomeFunit', None)
        self._check_end_of_publishes()

    def _set_enginery_guard(self, enginery_id, value, expected_unit=None, expected_value=None):
        e = self.adapter._entities[enginery_id]
        assert isinstance(e, Enginery)
        e.recipe._guarded = value
        self.sb.on_level_changed(e)
        if expected_unit:
            self._check_publish_state(expected_unit, expected_value)

        self._check_end_of_publishes()

    def test_fire_changing(self):
        self._set_enginery_guard(2, False, 'FireSensorsGuard', False)
        self._set_enginery_guard(0, True, 'FireSensorsGuard', None)
        self._set_enginery_guard(1, True)
        self._set_enginery_guard(2, True, 'FireSensorsGuard', True)

        self._set_enginery_guard(5, False, 'IntruderSensorsGuard', False)
        self._set_enginery_guard(8, False, 'LeakageSensorsGuard', False)

        self._set_enginery_guard(2, False, 'FireSensorsGuard', None)
        self._set_enginery_guard(0, False)
        self._set_enginery_guard(1, False, 'FireSensorsGuard', False)

    def test_intruder_changing(self):
        self._set_enginery_guard(5, False, 'IntruderSensorsGuard', False)
        self._set_enginery_guard(3, True, 'IntruderSensorsGuard', None)
        self._set_enginery_guard(4, True)
        self._set_enginery_guard(5, True, 'IntruderSensorsGuard', True)

        self._set_enginery_guard(2, False, 'FireSensorsGuard', False)
        self._set_enginery_guard(8, False, 'LeakageSensorsGuard', False)

        self._set_enginery_guard(5, False, 'IntruderSensorsGuard', None)
        self._set_enginery_guard(3, False)
        self._set_enginery_guard(4, False, 'IntruderSensorsGuard', False)

    def test_leakage_changing(self):
        self._set_enginery_guard(8, False, 'LeakageSensorsGuard', False)
        self._set_enginery_guard(6, True, 'LeakageSensorsGuard', None)
        self._set_enginery_guard(7, True)
        self._set_enginery_guard(8, True, 'LeakageSensorsGuard', True)

        self._set_enginery_guard(2, False, 'FireSensorsGuard', False)
        self._set_enginery_guard(5, False, 'IntruderSensorsGuard', False)

        self._set_enginery_guard(8, False, 'LeakageSensorsGuard', None)
        self._set_enginery_guard(6, False)
        self._set_enginery_guard(7, False, 'LeakageSensorsGuard', False)

    def _check_guard(self, funit, provider_ids):
        for value in True, False:
            funit(self.sb, Variable(value)).set()
            for provider_id in provider_ids:
                self._check_publish_command(provider_id, "Guard", value)

            self._check_end_of_publishes()

    def test_guard_setting(self):
        self._check_guard(alarming.Guard, [1000+i for i in range(9)])

    def test_fire_guard_setting(self):
        self._check_guard(alarming.FireSensorsGuard, [1000+i for i in range(3)])

    def test_intruder_guard_setting(self):
        self._check_guard(alarming.IntruderSensorsGuard, [1000+i for i in range(3, 6)])

    def test_leakage_guard_setting(self):
        self._check_guard(alarming.LeakageSensorsGuard, [1000+i for i in range(6, 9)])
