from functools import reduce
from typing import Generator

from django.contrib.auth.mixins import LoginRequiredMixin
from django.views.generic import TemplateView

import spread_core.mqtt as mqtt
from spread_core.bam import engineries
from spread_dashboard.project import bam
from spread_dashboard.responses import Error
from spread_dashboard.commissioning import models


class CommissioningIndex(LoginRequiredMixin, TemplateView):
    template_name = 'commissioning/index.html'

    def get(self, request, *args, **kwargs):
        return self.render_to_response({
            'title_1_href': f'commissioning/admin/',
            'locations': bam.models.Location.as_tree(),
            'projectName': bam.models.Header.objects.get().name,
            'broker': bam.broker_auth()
        })


class Intro(LoginRequiredMixin, TemplateView):
    template_name = 'commissioning/intro.html'


class Glossary(LoginRequiredMixin, TemplateView):
    template_name = 'commissioning/glossary.html'


class CommissioningLocation(LoginRequiredMixin, TemplateView):
    template_name = 'commissioning/location.html'

    def get(self, request, *args, location_id=None, **kwargs):
        try:
            loc = bam.models.Location.objects.get(id=location_id)

            navigations = (
                {
                    'id': parent_id,
                    'name': bam.models.Location.objects.get(id=parent_id).name
                }
                for parent_id in loc.navigations
            )

            areas = list(loc.engineries((engineries.LightingArea.__name__,)))

            try:
                strategy = reduce(
                    lambda a, b: set(a).union(b),
                    [
                        {key for key in s.keys() if s[key] and key not in ('id', 'area_id')}
                        for s in models.CommissioningStrategy.objects.filter(area__in=areas).values()
                    ]
                )
            except:
                strategy = []

            location = {
                'id': loc.id,
                'name': loc.name,
                'preview': f'commissioning/preview/{location_id}.jpg',
                'chart': location_chart_topic(loc),
                'locations': [
                    {
                        'id': sub_location.id,
                        'name': sub_location.name,
                        'actuators_count': enginery_count_on_location(sub_location, engineries.light_types),
                        'chart': location_chart_topic(sub_location)
                    }
                    for sub_location in loc.sub_locations
                ],
                'areas': [
                    {
                        'id': area.id,
                        'name': area.v_name,
                        'actuators': list(engineries_on_area(area, 'Actuation'))
                    } for area in areas
                ],
                'strategy': strategy,
                'actuators_count': enginery_count_on_location(loc, engineries.light_types),
                'presences_count': enginery_count_on_location(loc, (engineries.PresenceSensor.__name__,)),
                'luminosities_count': enginery_count_on_location(loc, (engineries.LightSensor.__name__,))
            }
            return self.render_to_response({
                'title_1_href': f'commissioning/admin/{location_id}/',
                'navigations': navigations,
                'location': location,
                'projectName': bam.models.Header.objects.get().name,
                'broker': bam.broker_auth()
            })

        except bam.models.Location.DoesNotExist:
            return Error('location <%s> does not exist' % location_id)


class CommissioningArea(LoginRequiredMixin, TemplateView):
    template_name = 'commissioning/area.html'

    def get(self, request, *args, location_id: int = None, area_id: int = None, **kwargs):
        try:
            loc = bam.models.Location.objects.get(id=location_id)
            area = bam.models.Enginery.objects.get(id=area_id)

            navigations = (
                *(
                    {
                        'id': parent_id,
                        'name': bam.models.Location.objects.get(id=parent_id).name
                    }
                    for parent_id in loc.navigations
                ),
                {
                    'id': area_id,
                    'name': area.v_name
                }
            )

            try:
                strategy = dict(
                    filter(
                        lambda item: item[1] is True,
                        models.CommissioningStrategy.objects.get(area=area).__dict__.items()
                    )
                ).keys()

            except models.CommissioningStrategy.DoesNotExist:
                strategy = ''

            return self.render_to_response({
                'title_1_href': f'commissioning/admin/{location_id}/{area_id}',
                'navigations': navigations,
                'area': {
                    'id': area_id,
                    'name': area.v_name,
                    'preview_lights': f'commissioning/preview/{location_id}_{area_id}_1.jpg',
                    'preview_sensors': f'commissioning/preview/{location_id}_{area_id}_2.jpg',
                    'actuators': list(engineries_on_area(area, 'Actuation')),
                    'presences': list(engineries_on_area(area, 'Luminosity')),
                    'luminosities': list(engineries_on_area(area, 'Presence')),
                    'buttons': list(engineries_on_area(area, 'Button')),
                    'state': area.state(''),
                    'strategy': strategy
                },
                'projectName': bam.models.Header.objects.get().name,
                'broker': bam.broker_auth()
            })

        except bam.models.Subginery.DoesNotExist:
            return Error('Subginery does not exist')
        except bam.models.Location.DoesNotExist:
            return Error('location <%s> does not exist' % location_id)


def enginery_count_on_location(location: bam.models.Location, engineries_types: (list, tuple)):
    count = len(list(location.engineries(engineries_types)))

    for sub_location in location.sub_locations:
        count += enginery_count_on_location(sub_location, engineries_types)

    return count


def engineries_on_area(area: bam.models.Enginery, ingredient_types: str) -> Generator[None, dict, None]:
    """
    Возвращает генератор инженерных объектов световой зоны.

    :param area: модель световой зоны
    :param ingredient_types: типы ингредиентов, которые должны попасть в выборку
    :return: Generator[None, dict, None]
    """
    for ing in area.recipe.get('ingredients', []):
        if ing['type'] == ingredient_types:
            try:
                enginery = bam.models.Enginery.objects.get(recipe={
                    "ingredients": [
                        {
                            "index": 0,
                            "type": "Solid",
                            "id": ing['id']
                        }
                    ],
                    "type": "Simple"
                })
            except bam.models.Enginery.DoesNotExist:
                continue
            else:
                try:
                    provider = bam.models.Provider.objects.get(id=ing['id'])
                    state = provider.state('')

                except (bam.models.Provider.DoesNotExist, bam.models.Manager.DoesNotExist):
                    state = None

                yield {
                    'id': enginery.id,
                    'state': state,
                    'chart': mqtt.spread.topic.Renew(
                        mqtt.spread.address.EngineryAddress(
                            bam.models.Header.objects.get().project_id,
                            enginery.type,
                            enginery.id,
                            'Chart'
                        )
                    ),
                    'power': enginery.attributes.get('power') if isinstance(enginery.attributes, dict) else None
                }


def location_chart_topic(location):
    try:
        return mqtt.spread.topic.Renew(
            mqtt.spread.address.SubgineryAddress(
                bam.models.Header.objects.get().project_id,
                bam.models.Subginery.objects.get(type='Lighting', locationID=location.id).id,
                'Lighting',
                'Chart'
            )
        )
    except bam.models.Subginery.DoesNotExist:
        return ''
